package com.tidspr.tsr;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.regex.Matcher;

import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import com.tidspr.tsr.entity.Diff;
import com.tidspr.tsr.entity.Tidder;
import com.tidspr.tsr.factory.URCF;
import com.tidspr.tsr.io.DiffIO;
import com.tidspr.tsr.io.TidderIO;
import com.tidspr.tsr.response.AbnormalPrinter;
import com.tidspr.tsr.response.AbnormalPrinter.ResCode;
import com.tidspr.tsr.security.Authority;
import com.tidspr.tsr.toolbox.Miscellaneous;
import com.tidspr.tsr.url.URLHandler;
import com.tidspr.tsr.url.URLHandler.PC;

import difflib.Delta;
import difflib.DiffUtils;
import difflib.Patch;

public class Wiki extends HttpServlet
{
	private static final long serialVersionUID = 1L;

	@Override
    public void doGet(final HttpServletRequest req, final HttpServletResponse res)
    {
    	if(!URLHandler.INSTANCE.handle(req, res, PC.Wiki)){
         	return;
        }
        wiki(req, res, new File(req.getPathInfo()).getName());
    }

    private static String _AuthLinkTemplate = "<div id='authlink'><ul><li>{{{AuthURL}}}</li><li>{{{UserName}}}</li></ul></div>";
    
    protected void wiki(final HttpServletRequest req, final HttpServletResponse res, final String s)
    {
    	HttpSession session = req.getSession(true);
        String id = (String)session.getAttribute("id");
    	Authority auth = (id!=null&&!id.isEmpty()) ? Authority.authorized:Authority.notauthorized;
    	
    	Tidder _or = TidderIO.readTidder(s);
        Tidder _pt = TidderIO.readTidder("PageTemplate");
        Tidder _mm = TidderIO.readTidder("MainMenu");

        if(_pt == null || _mm == null) return;
        
        final String auth_link =  _AuthLinkTemplate.replace("{{{UserName}}}", auth.isAuthorized() ? id:"")
        										   .replace("{{{AuthURL}}}", auth.isAuthorized()?"<a href='/logout?redirect=/wiki/"+URCF.encode(s)+"'>logout</a>":"<a href='/login?redirect=/wiki/"+URCF.encode(s)+"'>login</a>");

        final String enable = (auth.isAuthorized()) ? "" : "disabled='true'";
         
        String wikihtml = _or == null ? "" : _or.html();
        if (!wikihtml.isEmpty() && wikihtml.contains("<!--[inq[")) {
          Matcher m = Miscellaneous.INSTANCE.pIncludes.matcher(wikihtml);
          if (m.find()) {
            StringBuffer sb = new StringBuffer();
            do
            {
              Tidder tInclude = TidderIO.readTidder(wikihtml.substring(m.start(1), m.end(1)));
              m.appendReplacement(sb, tInclude == null ? "" : tInclude.html());
            }
            while (m.find());
            m.appendTail(sb);
            wikihtml = sb.toString();
          }
        }
        
        res.setContentType("text/html");
        res.setCharacterEncoding("UTF-8");
        try {
			final PrintWriter out = res.getWriter();
			out.write(
					_pt.raw().getValue()
					.replace("{{{MainMenu}}}",_mm.html())
					.replace("{{{auth_link}}}",auth_link)
					.replace("{{{wiki.title.decoded}}}",s)//not decoded actually
					.replace("{{{wiki.title.encoded}}}",URCF.encode(s))
					.replace("{{{enable}}}",enable)
					.replace("{{{modified}}}",(_or==null)?"":new SimpleDateFormat("dd MMMMM yyyy").format(_or.date()))
					.replace("{{{wiki.html}}}",(_or==null)?"":wikihtml.replace("{{{", "&#123;&#123;&#123;").replace("}}}", "&#125;&#125;&#125;"))
					.replace("{{{wiki.raw}}}",(_or==null)?"":_or.raw().getValue().replace("<", "&lt;").replace(">", "&gt;"))		
					);
		} catch (IOException e) {
			e.printStackTrace();
		}
    }
    
    
    @Override
    public void doPost(final HttpServletRequest req, final HttpServletResponse res)
    {
    	HttpSession session = req.getSession(true);
        String id = (String)session.getAttribute("id");
    	Authority auth = (id!=null&&!id.isEmpty()) ? Authority.authorized:Authority.notauthorized;
    	if(!auth.isAuthorized()){
    		AbnormalPrinter.Singleton.print(res, req, ResCode.Unauthorized);
    		return;
    	}

    	final String name = new File(req.getPathInfo()).getName();
        final String raw = req.getParameter("raw").replaceAll("\r\n?", "\n");

        if ((name == null) || name.isEmpty())
        {
        	AbnormalPrinter.Singleton.print(res,req,ResCode.BadRequest);
            return;
        }

        Tidder _or = TidderIO.readTidder(name);
        Diff diff_or = DiffIO.readDiff(name);
        
        final String _diff = (diff_or != null) ? diff_or.html() : "";
        final String _raw = (_or != null) ? _or.raw().getValue() : "";
        final List<String> original = Arrays.asList(_raw.split("\n"));
        final List<String> revised = Arrays.asList(raw.split("\n"));

    	StringBuilder sb = new StringBuilder();
        sb.append("<div class='date'>" + new SimpleDateFormat("yyyy.MM.dd kk:mm:ss z EEEEE").format(new Date()) + "</div>\n");
    	Patch patch = DiffUtils.diff(original, revised);
    	List<Delta> deltas = patch.getDeltas();
    	int last = 0;
    	for (Delta delta : deltas) {
    		if(last + 1 < delta.getOriginal().getPosition()){} // not continuous
		    List<?> or = delta.getOriginal().getLines();
    		List<?> re = delta.getRevised().getLines();
    		String orText = join(or).replace("<", "&lt;").replace(">", "&gt;");
    		String reText = join(re).replace("<", "&lt;").replace(">", "&gt;");
    		if(orText.trim().isEmpty()){
    			sb.append("<pre class='insert'>" + reText + "</pre>\n");
    		}else if(reText.trim().isEmpty()){
    			sb.append("<pre class='del'>" + orText + "</pre>\n");
    		}else{
                sb.append("<div class='mod'>\n");
                sb.append("\t<pre class='insert'>" + reText + "</pre>\n");
                sb.append("\t<pre class='del'>" + orText + "</pre>\n");
                sb.append("</div>\n");
    		}
			last = delta.getOriginal().last();
		}
		if(last + 1 < original.size()){}//last is not delta
        
		sb.append(_diff);
		
        // override
		TidderIO.writeTidder(name,raw);
        DiffIO.writeDiff(name,sb.toString());
    }
    
    private String join(List<?> l){
    	StringBuilder sb = new StringBuilder();
    	sb.delete(0, sb.length());
		for (Object object : l) {
			sb.append(object);
			sb.append("\n");
		}
		if(sb.length()>=1)sb.deleteCharAt(sb.length()-1); // last "\n"
		return sb.toString();
    }
}
